/* Copyright (C) 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package backend.srb;

import osio.TextPrinter;
import compbase.Mthd;
import compbase.StdTypes;
import backend.Instruction;
import backend.x86.IA32;

/**
 * SRBIA32: generate IA32 code through convertion of SRB blocks 
 * 
 * @author S. Frenz
 * @version 090306 fixed multi-setting of possible native register for virtual register
 *  version 090305 added support for binary operations
 *  version 090219 adopted changed X86Base
 *  version 090218 reduced access level of getReg
 *  version 090215 added first step towards native register allocation
 *  version 090207 added copyright notice
 *  version 090118 added getRegisterCount
 *  version 081226 initial version
 */

public class SRBT_IA32 extends IA32 implements SRBTarget {
  //our SRB master
  protected SRB srb;
  
  public void backInit(SRB is) {
    srb=is;
    is.relocBytes=relocBytes;
    is.allocClearBits=allocClearBits;
    is.stackClearBits=stackClearBits;
    is.maxInstrCodeSize=maxInstrCodeSize;
  }
  
  public boolean hasError() {
    return err;
  }
  
  public int getRegisterCount(int type) {
    switch (type) {
      case StdTypes.T_BOOL:
      case StdTypes.T_BYTE:
      case StdTypes.T_SHRT:
      case StdTypes.T_INT:
      case StdTypes.T_CHAR:
      case StdTypes.T_FLT:
      case StdTypes.T_DBL:
      case StdTypes.T_PTR:
        return 1;
      case StdTypes.T_LONG:
      case StdTypes.T_DPTR:
        return 2;
    }
    ctx.out.println("invalid type in getRegisterCount");
    err=true;
    return 0;
  }
  
  public SRBNativeRegisterInfo createNativeRegisterInfo() {
    return new SRBNativeRegisterInfo();
  }
  
  public void initSpecialRegister(SRBRegister reg) {
    int nReg;
    switch (reg.nr) {
      case SRB.R_CLSS: nReg=RegClss; break;
      case SRB.R_INST: nReg=RegInst; break;
      case SRB.R_BASE: nReg=RegBase; break;
      default:
        return; //not a special register, leave untouched
    }
    reg.nativeInfo.nativeReg=reg.nativeInfo.completeNativeReg=nReg;
  }
  
  public void resolveRegister(SRBRegister reg) {
    SRBConstraint cst=null;
    SRBNativeRegisterInfo otherInfo;
    boolean registerRequired=false;
    if (reg.nr<SRB.R_GPRS) return; //already initialized in initSpecialRegister
    //fulfill constraints
    while ((cst=srb.getNextConstraint(SRBConstraint.C_ALL, reg, cst))!=null) {
      switch (cst.type) {
        case SRBConstraint.C_EQUALS:
          otherInfo=getOtherReg(reg, cst.reg1, cst.reg2).nativeInfo;
          if (otherInfo.nativeReg!=0) reg.nativeInfo.possibleNativeReg=otherInfo.nativeReg;
          break;
        case SRBConstraint.C_AVOIDS:
          reg.nativeInfo.possibleNativeReg&=~getOtherReg(reg, cst.reg1, cst.reg2).nativeInfo.nativeReg;
          break;
        case SRBConstraint.C_REQUIRED:
          registerRequired=true;
          break;
      }
    }
    //constraints fulfilled, fix native register
    if (registerRequired) {
      if (reg.nativeInfo.possibleNativeReg==0) {
        ctx.out.print("could not resolve register ");
        ctx.out.println(reg.nr);
        err=true;
      }
      else reg.nativeInfo.nativeReg=getFirstSetBit(reg.nativeInfo.possibleNativeReg);
    }
    else reg.nativeInfo.nativeReg=0;
  }
  
  public void mapNativeRegister(SRBRegister reg) {
    //merge multiple-registers
    if (reg.type==SRB.T_SPRT) {
      reg.nativeInfo.completeNativeReg=0; //do not access supporting register directly
      reg.containerPrev.nativeInfo.completeNativeReg|=reg.nativeInfo.nativeReg; //activate register mask of supporting register in master register
    }
    else reg.nativeInfo.completeNativeReg=reg.nativeInfo.nativeReg;
  }
  
  private int getFirstSetBit(int mask) {
    int i=1;
    while (i!=0) {
      if ((mask&i)!=0) return i;
      i=i<<1;
    }
    ctx.out.println("no set bit found in getFirstSetBit");
    err=true;
    return 0;
  }
  
  private SRBRegister getOtherReg(SRBRegister search, SRBRegister reg1, SRBRegister reg2) {
    if (search==reg1 && reg2!=null) return reg2;
    if (search==reg2 && reg1!=null) return reg1;
    ctx.out.println("no other reg found in getOtherReg");
    err=true;
    return search; //avoid null pointer
  }
  
  public void printNativeInfo(SRBNativeRegisterInfo info, TextPrinter v) {
    v.print("0x");
    v.printHexFix(info.completeNativeReg, 4);
  }
  
  //implemented methods for SRBTarget
  
  public void handleMthdFrame(int enter, Mthd refMthd, int phase) {
    switch (phase) {
      case SRB.P_GENERATECODE:
        if (enter==1) codeProlog(refMthd);
        else codeEpilog(refMthd);
    }
  }
  
  public void handleFlowHint(int hint, int id, int phase) {
    //insertFlowHint(hint, id);
  }
  
  public void handleLoadConstVal(SRBRegister dst, int ival, long lval, int type, int phase) {
    boolean asDouble=false;
    switch (phase) {
      case SRB.P_DECORATEREGISTERS:
        switch (type) {
          case StdTypes.T_BOOL:
          case StdTypes.T_BYTE:
          case StdTypes.T_CHAR:
          case StdTypes.T_SHRT:
          case StdTypes.T_INT:
          case StdTypes.T_PTR:
            dst.nativeInfo.possibleNativeReg=RegA|RegB|RegC|RegD;
            srb.addConstraint(SRBConstraint.C_CONSTINT, dst, ival, false);
            break;
          case StdTypes.T_LONG:
          case StdTypes.T_DPTR:
            dst.nativeInfo.possibleNativeReg=RegA|RegB;
            dst.containerNext.nativeInfo.possibleNativeReg=RegD|RegC;
            srb.addConstraint(SRBConstraint.C_CONSTLONG, dst, lval, true);
            break;
        }
        break;
      case SRB.P_GENERATECODE:
        switch (type) {
          case StdTypes.T_DBL:
            asDouble=true;
          case StdTypes.T_LONG:
            if (srb.getNextConstraint(SRBConstraint.C_REQUIRED, dst, null)!=null)
              genLoadConstDoubleOrLongVal(dst.nativeInfo.completeNativeReg, lval, asDouble);
            break;
          default:
            if (srb.getNextConstraint(SRBConstraint.C_REQUIRED, dst, null)!=null)
              genLoadConstVal(dst.nativeInfo.completeNativeReg, ival, type);
        }
        break;
    }
  }
  
  public void handleLoadVar(boolean loadValueNotAddress, SRBRegister dst, SRBRegister src, int pos, int type, int phase) {
    switch (phase) {
      case SRB.P_DECORATEREGISTERS:
        if (type==StdTypes.T_LONG || type==StdTypes.T_DPTR) {
          dst.nativeInfo.possibleNativeReg=RegA|RegB|RegClss|RegInst;
          dst.containerNext.nativeInfo.possibleNativeReg=RegC|RegD;
        }
        else dst.nativeInfo.possibleNativeReg=RegA|RegB|RegC|RegD|RegClss|RegInst;
        if (src==null) {
          srb.addConstraint(SRBConstraint.C_CONSTINT, dst, pos, false);
        }
        break;
      case SRB.P_GENERATECODE:
        if (srb.getNextConstraint(SRBConstraint.C_REQUIRED, dst, null)!=null) {
          if (loadValueNotAddress) { //load address
            genLoadVarVal(dst.nativeInfo.completeNativeReg, src.nativeInfo.completeNativeReg, src==null ? M_ABSL : M_RESU, pos, type);
          }
          else { //load value
            if (src==null) genLoadConstVal(dst.nativeInfo.completeNativeReg, pos, StdTypes.T_PTR); //absolute address stored in pos
            else genLoadVarAddr(dst.nativeInfo.completeNativeReg, src.nativeInfo.completeNativeReg, M_RESU, pos);
          }
        }
        break;
    }
  }
  
  public void handleDupConv(SRBRegister dst, SRBRegister src, int toType, int fromType, boolean intfMapSpecial, int phase) {
    switch (phase) {
      case SRB.P_DECORATEREGISTERS:
        if (intfMapSpecial) {
          if (dst.type!=StdTypes.T_DPTR || src.nr!=SRB.R_PRIR) {
            ctx.out.println("invalid intfMapSpecial in handleDupConv");
            err=true;
          }
        }
        else if (src.nr==SRB.R_PRIR) { //source is primary => map destination to correct register
          dst.nativeInfo.possibleNativeReg&=RegA;
          if (dst.allocNext!=null && dst.allocNext.type==SRB.T_SPRT)
            dst.allocNext.nativeInfo.possibleNativeReg&=RegD;
        }
        else if (dst.nr==SRB.R_PRIR) { //destination is primary => map source to correct register
          src.nativeInfo.possibleNativeReg&=RegA;
          if (src.allocNext!=null && src.allocNext.type==SRB.T_SPRT)
            src.allocNext.nativeInfo.possibleNativeReg&=RegD;
        }
        else { //all registers possible
          dst.nativeInfo.possibleNativeReg=RegA|RegB|RegC|RegD|RegClss|RegInst;
        }
        break;
      case SRB.P_GENERATECODE:
        if (intfMapSpecial) genMoveIntfMapFromPrimary(dst.nativeInfo.completeNativeReg);
        else if (dst.nr!=SRB.R_PRIR) {
          if (fromType!=toType) genConvertVal(dst.nativeInfo.completeNativeReg, src.nativeInfo.completeNativeReg, toType, fromType);
          else if (dst!=src) genDup(dst.nativeInfo.completeNativeReg, src.nativeInfo.completeNativeReg, fromType);
          //else: nothing to do
        }
        //else: do not insert move to/from primary result register, should already be in destination due to register mapping constraint
        break;
    }
  }
  
  public void handleSingleStackOp(boolean push, SRBRegister reg, int ival, long lval, int type, int phase) {
    //TODO
  }
  
  public void handleAssign(SRBRegister dst, SRBRegister src, int type, int phase) {
    SRBConstraint constSrc, constDst;
    switch (phase) {
      case SRB.P_DECORATEREGISTERS:
        if (srb.getNextConstraint(SRBConstraint.C_CONSTINT, src, null)!=null) {
          if (srb.getNextConstraint(SRBConstraint.C_CONSTINT, dst, null)==null)
            srb.addConstraint(SRBConstraint.C_REQUIRED, dst, true);
          //else: no register required
        }
        else {
          srb.addConstraint(SRBConstraint.C_REQUIRED, src, true);
          srb.addConstraint(SRBConstraint.C_REQUIRED, dst, false);
          srb.addConstraint(SRBConstraint.C_AVOIDS, dst, src, true);
        }
        break;
      case SRB.P_GENERATECODE:
        if ((constSrc=srb.getNextConstraint(SRBConstraint.C_CONSTINT, src, null))!=null) {
          if ((constDst=srb.getNextConstraint(SRBConstraint.C_CONSTINT, dst, null))!=null)
            genStoreVarConstVal(0, constSrc.constInt, M_ABSL, constDst.constInt, type);
          else genStoreVarConstVal(dst.nativeInfo.completeNativeReg, constSrc.constInt, M_RESU, 0, type);
        }
        else genAssign(dst.nativeInfo.completeNativeReg, src.nativeInfo.completeNativeReg, type);
        break;
    }
  }
  
  public void handleRegOp(SRBRegister dst, SRBRegister src1, SRBRegister src2, int op, int type, int phase) {
    switch (phase) {
      case SRB.P_DECORATEREGISTERS:
        srb.addConstraint(SRBConstraint.C_REQUIRED, dst, true);
        srb.addConstraint(SRBConstraint.C_REQUIRED, src1, true);
        srb.addConstraint(SRBConstraint.C_REQUIRED, src2, true);
        srb.addConstraint(SRBConstraint.C_EQUALS, dst, src1, true);
        srb.addConstraint(SRBConstraint.C_AVOIDS, src1, src2, true);
        if (type==StdTypes.T_LONG || type==StdTypes.T_DPTR) {
          dst.nativeInfo.possibleNativeReg=RegA|RegB;
          dst.containerNext.nativeInfo.possibleNativeReg=RegC|RegD;
        }
        else dst.nativeInfo.possibleNativeReg=RegA|RegB|RegC|RegD;
        break;
      case SRB.P_GENERATECODE:
        genBinOp(dst.nativeInfo.completeNativeReg,
            src1.nativeInfo.completeNativeReg, src2.nativeInfo.completeNativeReg, op, type);
        break;
    }
  }
  
  public void handleMemOp(SRBRegister dst, int op, int type, int phase) {
    //TODO
  }
  
  public void handleCall(SRBRegister baseReg, int off, int parSize, Mthd mthd, boolean withIntfMap, int phase) {
    switch (phase) {
      case SRB.P_DECORATEREGISTERS:
        if (mthd==null) srb.addConstraint(SRBConstraint.C_REQUIRED, baseReg, true);
        break;
      case SRB.P_GENERATECODE:
        if (mthd!=null) genCallConst(mthd, parSize);
        else if (withIntfMap) genCallIndexed(baseReg.nativeInfo.completeNativeReg, off, parSize);
        else if (baseReg.nr==SRB.R_CLSS) genCall(off, 0, true, parSize);
        else genCall(off, baseReg.nativeInfo.completeNativeReg, false, parSize);
        break;
    }
  }
  
  public void handleJump(Instruction dest, int cond, int phase) {
    switch (phase) {
      case SRB.P_GENERATECODE:
        insJump(dest, cond);
        break;
    }
  }
  
  public void handleCompare(SRBRegister src1, SRBRegister src2, int cond, int val, long lval, int type, int phase) {
    //TODO
  }
  
  public void handleSpecialMemIO(boolean write, SRBRegister dst, SRBRegister src, int type, int memLoc, int phase) {
    //TODO
  }
  
  public void handleDerefAddr(SRBRegister addrReg, SRBRegister indReg, int baseOffset, int entrySize, int phase) {
    //TODO
  }
  
  public void handleCheckBounds(SRBRegister addrReg, SRBRegister offReg, int checkToOffset, Instruction onSuccess, int phase) {
    //TODO
  }
  
  public void handleNativeBoundException(int phase) {
    switch (phase) {
      case SRB.P_GENERATECODE:
        genNativeBoundException();
        break;
    }
  }
  
  public void handleThrowFrame(int iNr, SRBRegister addrReg, Instruction dest, int throwBlockOffset, int phase) {
    //TODO
  }
  
  public void handleInlineVarOffset(int inlineMode, SRBRegister src, int pos, int baseValue, int phase) {
    //TODO
  }
}
