/* Copyright (C) 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package symbols;

import osio.TextBuffer;
import osio.TextPrinter;
import compbase.*;

/**
 * RawSymbols: generate output for all packages, append information at units and methods
 * 
 * @author S. Frenz
 * @version 090207 added copyright notice
 *  version 080701 adopted move of TextBuffer to osio
 *  version 080614 adopted changed Unit.searchVariable
 *  version 080115 adopted changes in SymbolInformer.generateSymbols, using SymbolInformer.checkVrbl
 *  version 080114 removed checkParameter
 *  version 070729 removed no longer needed prepareUnits, added type check for root-variable
 *  version 070727 added flag for special (String,int)-method-detection
 *  version 070714 initial version
 */

public class RTESymbols extends SymbolInformer {
  private Unit packUnit;
  private TextBuffer myBuffer;
  private int packNamePos, packOuterPos, packSubPacksPos, packNextPackPos, packUnitsPos;
  private int unitNamePos, unitModifierPos, unitNextUnitPos, unitPackPos, unitMthdsPos;
  private int mthdNameParPos, mthdModifierPos, mthdNextMthdPos;
  private int addedUnitSize, addedMthdSize;
  
  public static void printValidParameters(TextPrinter v) {
    v.println("   --- no parameter required");
  }
  
  public boolean setParameter(String parm, TextPrinter v) {
    v.println("No parameter expected for RTESymbols");
    return false;
  }

  public boolean generateSymbols(UnitList newUnits, Context ictx) {
    int rootAddr;
    Vrbl rootVar;
    
    ctx=ictx;
    myBuffer=new TextBuffer();
    //allocate root-package-object
    if ((packUnit=ctx.rte.searchUnit("SPackage"))==null
        || (packUnit.modifier&(Unit.M_ABSTR|Unit.M_INDIR|Unit.M_STRUCT))!=0) {
      ctx.out.println("unit rte.Package required for RTESymbols (non-abstract, non-indir, non-struct)");
      return false;
    }
    if ((rootVar=packUnit.searchVariable("root", ctx))==null
        || (rootVar.modifier&(Vrbl.M_STAT|Vrbl.M_STRUCT))!=Vrbl.M_STAT
        || rootVar.type.baseType!=TypeRef.T_QID || rootVar.type.qid.unitDest!=packUnit
        || rootVar.type.arrDim!=0) {
      ctx.out.println("variable static SPackage root needed in rte.SPackage");
      return false;
    }
    if ((packNamePos=checkVrbl(packUnit, "name"))==-1
        || (packOuterPos=checkVrbl(packUnit, "outer"))==-1
        || (packSubPacksPos=checkVrbl(packUnit, "subPacks"))==-1
        || (packNextPackPos=checkVrbl(packUnit, "nextPack"))==-1
        || (packUnitsPos=checkVrbl(packUnit, "units"))==-1
        || (unitNamePos=checkVrbl(ctx.rteSClassDesc, "name"))==-1
        || (unitModifierPos=checkVrbl(ctx.rteSClassDesc, "modifier"))==-1
        || (unitNextUnitPos=checkVrbl(ctx.rteSClassDesc, "nextUnit"))==-1
        || (unitPackPos=checkVrbl(ctx.rteSClassDesc, "pack"))==-1
        || (unitMthdsPos=checkVrbl(ctx.rteSClassDesc, "mthds"))==-1
        || (mthdNameParPos=checkVrbl(ctx.rteSMthdBlock, "namePar"))==-1
        || (mthdNextMthdPos=checkVrbl(ctx.rteSMthdBlock, "nextMthd"))==-1
        || (mthdModifierPos=checkVrbl(ctx.rteSMthdBlock, "modifier"))==-1) {
      return false;
    }
    //calculate size of fields required for symbol information
    addedUnitSize=(4*ctx.arch.relocBytes+4+ctx.arch.allocClearBits)&~ctx.arch.allocClearBits;
    addedMthdSize=(2*ctx.arch.relocBytes+4+ctx.arch.allocClearBits)&~ctx.arch.allocClearBits;
    //generate tree and enter root-object in static root-field
    if ((rootAddr=genPackage(ctx.root, 0))==0) return false;
    ctx.arch.putRef(ctx.dynaMem ? packUnit.outputAddr-rootVar.varPos*ctx.arch.relocBytes
        : rootVar.varPos, rootAddr);
    //everything done
    return !ctx.err;
  }
  
  private int genPackage(Pack pack, int outerAddr) {
    int firstPackAddr=0, packAddr=0, lastPackAddr=0, nameAddr;
    
    while (pack!=null) {
      if ((packAddr=ctx.mem.allocate(packUnit.instScalarTableSize, packUnit.instIndirScalarTableSize,
          packUnit.instRelocTableEntries, packUnit.outputAddr))<=0) {
        ctx.out.println("error allocating package");
        ctx.err=true;
        return -1;
      }
      if (pack.name!=null) {
        if ((nameAddr=ctx.allocateString(pack.name))<=0) return -1;
        ctx.arch.putRef(packAddr-packNamePos, nameAddr);
      }
      ctx.arch.putRef(packAddr-packOuterPos, outerAddr);
      if (firstPackAddr==0) firstPackAddr=packAddr;
      if (lastPackAddr!=0) ctx.arch.putRef(lastPackAddr-packNextPackPos, packAddr);
      ctx.arch.putRef(packAddr-packUnitsPos, genUnit(pack.units, packAddr));
      ctx.arch.putRef(packAddr-packSubPacksPos, genPackage(pack.subPacks, packAddr));
      lastPackAddr=packAddr;
      pack=pack.nextPack;
    }
    return firstPackAddr;
  }
  
  private int genUnit(Unit unit, int packAddr) {
    int firstUnitAddr=0, lastUnitAddr=0;
    
    while (unit!=null) {
      if (unit.outputAddr!=0 && (unit.modifier&Unit.M_INDIR)==0) { //skip dummy-units of compiler and not-generated descriptors
        ctx.symGenSize+=addedUnitSize;
        if (firstUnitAddr==0) firstUnitAddr=unit.outputAddr;
        if (lastUnitAddr!=0) ctx.arch.putRef(lastUnitAddr-unitNextUnitPos, unit.outputAddr);
        lastUnitAddr=unit.outputAddr;
        ctx.arch.putRef(unit.outputAddr-unitNamePos, ctx.allocateString(unit.name));
        ctx.arch.putRef(unit.outputAddr-unitPackPos, packAddr);
        ctx.arch.putRef(unit.outputAddr-unitMthdsPos, genMthds(unit.mthds));
        ctx.arch.putInt(unit.outputAddr+unitModifierPos, unit.modifier);
      }
      unit=unit.nextUnit;
    }
    return firstUnitAddr;
  }
  
  private int genMthds(Mthd mthd) {
    int firstMthdAddr=0, lastMthdAddr=0, command;
    
    while (mthd!=null) {
      if (mthd.outputAddr!=0) { //skip not generated methods
        ctx.symGenSize+=addedMthdSize;
        if (firstMthdAddr==0) firstMthdAddr=mthd.outputAddr;
        if (lastMthdAddr!=0) ctx.arch.putRef(lastMthdAddr-mthdNextMthdPos, mthd.outputAddr);
        lastMthdAddr=mthd.outputAddr;
        myBuffer.used=0;
        mthd.printNamePar(myBuffer);
        ctx.arch.putRef(mthd.outputAddr-mthdNameParPos,
            ctx.allocateString(new String(myBuffer.data, 0, myBuffer.used)));
        if (mthd.param!=null &&
            mthd.param.type.compareType(ctx.stringType, true, ctx)!=TypeRef.C_NP
            && mthd.param.nextParam!=null && mthd.param.nextParam.type.isIntType()
            && mthd.param.nextParam.nextParam==null) command=0x80000000;
        else command=0;
        ctx.arch.putInt(mthd.outputAddr+mthdModifierPos, mthd.modifier|command);
      }
      mthd=mthd.nextMthd;
    }
    return firstMthdAddr;
  }
}
